/*
 *  Copyright (C) 2011 Masaryk University, Faculty of Informatics
 *
 *  Author: Daniel Kopecek <xkopecek@fi.muni.cz>
 *
 *  LICENSE TERMS
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions and the following disclaimer in
 *  the documentation and/or other materials provided with the
 *  distribution.
 *  3. Neither the name of the Company nor the names of its contributors
 *  may be used to endorse or promote products derived from this
 *  software without specific prior written permission.
 *
 *  This software is provided ``as is'', and any express or implied
 *  warranties, including, but not limited to, the implied warranties of
 *  merchantability and fitness for a particular purpose are disclaimed.
 *  In no event shall the company or contributors be liable for any
 *  direct, indirect, incidental, special, exemplary, or consequential
 *  damages (including, but not limited to, procurement of substitute
 *  goods or services; loss of use, data, or profits; or business
 *  interruption) however caused and on any theory of liability, whether
 *  in contract, strict liability, or tort (including negligence or
 *  otherwise) arising in any way out of the use of this software, even
 *  if advised of the possibility of such damage.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <unistd.h>
#include <errno.h>
#include <libgen.h>
#include <libconfig.h>
#include <libtriplex/libtriplex.h>
#include <libtriplex/gff3.h>

#include "common/helpers.h"

const char short_options[] = "P:C:o:s:e:l:L:m:i:j:t:p:a:b:c:d:f:ISJ:K:D:M:N:vh";

static bool verbose_mode = false;

static int steps, min_score, min_length, min_loop_length;
static int max_loop_length, max_length, max_segment;

static int pen_isochange, pen_mismatch, pen_insdel;
static int dtwist_limit, score_isostay;

static int *tri_types = NULL;
static size_t tri_count = 0;

static double p_val;
static double lambda0145, lambda2367;
static double mi0145, mi2367;

void print_usage(FILE *fp, triplexCTX *ctx, char *arg0)
{
        register size_t i;
        (void)ctx;

        fprintf(fp, "\n");
        fprintf(fp, " Usage: %s [OPTIONS] <input-file>\n\n", basename(arg0));

        fprintf(fp, "\t-C <path>         Path to the configuration file (default: %s)\n", "none");
        fprintf(fp, "\t-P <string>       Configuration profile\n");
        fprintf(fp, "\t-f <path>         Input file.\n");
        fprintf(fp, "\t-o <path>         Output path\n");
        fprintf(fp, "\t-I                Turn off FASTA indexing.\n");
        fprintf(fp, "\t-S                Turn off CDS mapping.\n");
        fprintf(fp, "\t-v                Verbose output.\n");
        fprintf(fp, "\t-h                Show this text.\n");
        fprintf(fp, "\n");
        fprintf(fp, "\t == Search parameters ==\n\n");
        fprintf(fp, "\t-s <int>          Number of steps (default: %u)\n", steps);
        fprintf(fp, "\t-e <int>          Minimal score treshold (default: %u)\n", min_score);
        fprintf(fp, "\t-l <int>          Minimal triplex length (default: %u)\n", min_length);
        fprintf(fp, "\t-L <int>          Maximal triplex length (default: %u)\n", max_length);
        fprintf(fp, "\t-m <int>          Maximal segment size (default: %u)\n", max_segment);
        fprintf(fp, "\t-l <int>          Minimal triplex length (default: %u)\n", min_length);
        fprintf(fp, "\t-i <int>          Minimal loop length (default: %u)\n", min_loop_length);
        fprintf(fp, "\t-j <int>          Maximal loop length (default: %u)\n", max_loop_length);

        fprintf(fp, "\t-t <int>[,...]    Triplex type(s) (default: ");
        for (i = 0; i < tri_count; ++i)
                fprintf(fp, "%u%c", tri_types[i], i + 1 == tri_count ? '\0' : ',');
        fprintf(fp, ").\n");

        fprintf(fp, "\t-p <float>        Acceptable P-value (default: %1.2f)\n", p_val);
        fprintf(fp, "\n");

        fprintf(fp, "\t == P-value calculation parameters ==\n\n");
        fprintf(fp, "\t-a <float>        Lambda for type 0,1,4,5 (default: %1.2f)\n", lambda0145);
        fprintf(fp, "\t-b <float>        Lambda for type 2,3,6,7 (default: %1.2f)\n", lambda2367);

        fprintf(fp, "\t-c <float>        Mi for type 0,1,4,5 (default: %1.2f)\n", mi0145);
        fprintf(fp, "\t-d <float>        Mi for type 2,3,6,7 (default: %1.2f)\n", mi2367);
        fprintf(fp, "\n");
        fprintf(fp, "\t == Scoring scheme ==\n\n");
        fprintf(fp, "\t-J <int>          Isogroup change penalization (default: %d)\n", pen_isochange);
        fprintf(fp, "\t-K <int>          Mismatch penalization (default: %d)\n", pen_mismatch);
        fprintf(fp, "\t-D <int>          Insdel penalization (default: %d)\n", pen_insdel);
        fprintf(fp, "\t-M <int>          dtwist limit (default: %d)\n", dtwist_limit);
        fprintf(fp, "\t-N <int>          Same isogroup increment (default: %d)\n", score_isostay);
        fprintf(fp, "\n");
}

/**
 * Read each all key=value pairs from the given profile and try to set
 * a context parameter using `key' as the parameter name and `value' as
 * its value.
 */
static int load_profile(triplexCTX *ctx, config_t *cfg, const char *profile)
{
        register int i, ret;
        config_setting_t *group, *param;

        group = config_lookup(cfg, profile);

        if (group == NULL) {
                fprintf(stderr, "[!] the configuration does not contain a profile named \"%s\"\n", profile);
                return (-1);
        }

        if (config_setting_is_group(group) != CONFIG_TRUE) {
                fprintf(stderr, "[!] the provided profile \"%s\" does not represent a profile in the configuration file \"%s\"\n",
                        profile, config_setting_source_file(group));
                return (-1);
        }

        i = 0;

        while ((param = config_setting_get_elem(group, i)) != NULL) {
                switch(config_setting_type(param)) {
                case CONFIG_TYPE_INT:
                        if ((ret = libtriplex_setparam(ctx, config_setting_name(param),
                                                       TRIPLEX_PARAM_INT, config_setting_get_int(param))) != 0)
                                goto failure;
                        break;
                case CONFIG_TYPE_FLOAT:
                        if ((ret = libtriplex_setparam(ctx, config_setting_name(param),
                                                       TRIPLEX_PARAM_FLOAT, config_setting_get_float(param))) != 0)
                                goto failure;
                        break;
                case CONFIG_TYPE_STRING:
                        if ((ret = libtriplex_setparam(ctx, config_setting_name(param),
                                                       TRIPLEX_PARAM_STRING, config_setting_get_string(param))) != 0)
                                goto failure;
                        break;
                case CONFIG_TYPE_ARRAY: {
                        int count = config_setting_length(param), ai;
                        config_setting_t *elem = config_setting_get_elem(param, 0);

                        switch(config_setting_type(elem)) {
                        case CONFIG_TYPE_INT: {
                                /*
                                 * Allocate a new array since the pointer is saved
                                 * in the triplexParam structure. It will be freed
                                 * by the libtriplex_fini function.
                                 */
                                int *array = alloc_array(int, count);

                                array[0] = config_setting_get_int(elem);

                                for (ai = 1; ai < count; ++ai)
                                        array[ai] = config_setting_get_int_elem(param, ai);

                                if ((ret = libtriplex_setparam(ctx, config_setting_name(param),
                                                               TRIPLEX_PARAM_INT|TRIPLEX_PFLAG_ARRAY,
                                                               array, count)) != 0)
                                        goto failure;
                        }       break;
                        case CONFIG_TYPE_FLOAT: {
                                double *array = alloc_array(double, count);
                        }       break;
                        case CONFIG_TYPE_STRING: {
                                char **string = alloc_array(char *, count);
                        }       break;
                        }

                }       break;
                failure:
                        fprintf(stderr, "[!] An error ocured while processing parameter `%s' in profile `%s': ",
                                config_setting_name(param), profile);

                        switch(ret) {
                        case LIBTRIPLEX_ESYSTEM:
                                fprintf(stderr, "system error: %s (errno=%d)\n", strerror(errno), errno);
                                break;
                        case LIBTRIPLEX_EINVALP:
                                fprintf(stderr, "unknown parameter\n");
                                break;
                        case LIBTRIPLEX_EINVALV:
                                fprintf(stderr, "invalid parameter value/type\n");
                                break;
                        default:
                                fprintf(stderr, "unknown error (probably a bug, please report this to the authors)\n");
                                break;
                        }

                        return (-1);
                default:
                        dP("Setting type %u not supported: %s\n", config_setting_type(param), config_setting_name(param));
                }

                ++i;
        }

        return (0);
}

int main(int argc, char *argv[])
{
        int ret = EXIT_SUCCESS;
        char *arg0 = argv[0];

        config_t    cfg;
        bool        cfg_loaded = false;
        triplexCTX *ctx;

        int opt;
        char *opt_profile  = NULL, *opt_outpath  = NULL;
        char *infile = NULL;

        gff3_t *gffres = NULL;

        config_init(&cfg);
        ctx = libtriplex_init();

        /*
         * Load builtin parameters first.
         */
        libtriplex_getparam(ctx, "triplex-type", TRIPLEX_PARAM_INT|TRIPLEX_PFLAG_ARRAY, &tri_types, &tri_count);
        libtriplex_getparam(ctx, "max-steps", TRIPLEX_PARAM_INT, &steps);
        libtriplex_getparam(ctx, "min-length", TRIPLEX_PARAM_INT, &min_length);
        libtriplex_getparam(ctx, "max-length", TRIPLEX_PARAM_INT, &max_length);
        libtriplex_getparam(ctx, "min-score", TRIPLEX_PARAM_INT, &min_score);
        libtriplex_getparam(ctx, "min-loop-length", TRIPLEX_PARAM_INT, &min_loop_length);
        libtriplex_getparam(ctx, "max-loop-length", TRIPLEX_PARAM_INT, &max_loop_length);
        libtriplex_getparam(ctx, "lambda_0145", TRIPLEX_PARAM_FLOAT, &lambda0145);
        libtriplex_getparam(ctx, "lambda_2367", TRIPLEX_PARAM_FLOAT, &lambda2367);
        libtriplex_getparam(ctx, "mi_0145", TRIPLEX_PARAM_FLOAT, &mi0145);
        libtriplex_getparam(ctx, "mi_2367", TRIPLEX_PARAM_FLOAT, &mi2367);

        libtriplex_getparam(ctx, "pen-isochange", TRIPLEX_PARAM_INT, &pen_isochange);
        libtriplex_getparam(ctx, "pen-mismatch",  TRIPLEX_PARAM_INT, &pen_mismatch);
        libtriplex_getparam(ctx, "pen-insdel",    TRIPLEX_PARAM_INT, &pen_insdel);
        libtriplex_getparam(ctx, "dtwist-limit",  TRIPLEX_PARAM_INT, &dtwist_limit);
        libtriplex_getparam(ctx, "score-isostay", TRIPLEX_PARAM_INT, &score_isostay);

        libtriplex_getparam(ctx, "p-value", TRIPLEX_PARAM_FLOAT, &p_val);
        libtriplex_getparam(ctx, "max-segment", TRIPLEX_PARAM_INT, &max_segment);

        /**
         * Process command line options
         */
        while ((opt = getopt(argc, argv, short_options)) != -1) {
                switch(opt) {
                case 'P':
                        if (!cfg_loaded) {
                                fprintf(stderr, "[!] no configuration loaded! Use the `-C' option before `-P' to load a configuration file\n");
                                ret = EXIT_FAILURE;
                                goto cleanup_and_exit;
                        }

                        if (load_profile(ctx, &cfg, optarg) != 0) {
                                ret = EXIT_FAILURE;
                                goto cleanup_and_exit;
                        }

                        break;
                case 'C':
                        /**
                         * Process configuration file
                         */
                        if (config_read_file(&cfg, optarg) != CONFIG_TRUE) {
                                switch(config_error_type(&cfg)) {
                                case CONFIG_ERR_FILE_IO:
                                        fprintf(stderr, "[!] unable to open/read the configuration file \"%s\"\n", optarg);
                                        break;
                                case CONFIG_ERR_PARSE:
                                        fprintf(stderr, "[!] parsing of configuration file \"%s\" failed at line %u: %s\n",
                                                optarg, config_error_line(&cfg), config_error_text(&cfg));
                                        break;
                                default:
                                        fprintf(stderr, "[!] config_read_file: unknown error\n");
                                }

                                ret = EXIT_FAILURE;
                                goto cleanup_and_exit;
                        }

                        if (load_profile(ctx, &cfg, "default") != 0) {
                                ret = EXIT_FAILURE;
                                goto cleanup_and_exit;
                        }

                        cfg_loaded = true;
                        break;
                case 'o':
                        opt_outpath = strdup(optarg);
                        break;
                case 's':
                        if ((ret = libtriplex_setparam_coerce(ctx, "max-steps", optarg)) != 0)
                                goto failure;
                        break;
                case 'e':
                        if ((ret = libtriplex_setparam_coerce(ctx, "min-score", optarg)) != 0)
                                goto failure;
                        break;
                case 'l':
                        if ((ret = libtriplex_setparam_coerce(ctx, "min-length", optarg)) != 0)
                                goto failure;
                        break;
                case 'L':
                        if ((ret = libtriplex_setparam_coerce(ctx, "max-length", optarg)) != 0)
                                goto failure;
                        break;
                case 'm':
                        if ((ret = libtriplex_setparam_coerce(ctx, "max-segment", optarg)) != 0)
                                goto failure;
                        break;
                case 'i':
                        if ((ret = libtriplex_setparam_coerce(ctx, "min-loop-length", optarg)) != 0)
                                goto failure;
                        break;
                case 'j':
                        if ((ret = libtriplex_setparam_coerce(ctx, "max-loop-length", optarg)) != 0)
                                goto failure;
                        break;
                case 't':
                        if ((ret = libtriplex_setparam_coerce(ctx, "triplex-type", optarg)) != 0)
                                goto failure;
                        break;
                case 'p':
                        if ((ret = libtriplex_setparam_coerce(ctx, "p-value", optarg)) != 0)
                                goto failure;
                        break;
                case 'a':
                        if ((ret = libtriplex_setparam_coerce(ctx, "lambda_0145", optarg)) != 0)
                                goto failure;
                        break;
                case 'b':
                        if ((ret = libtriplex_setparam_coerce(ctx, "lambda_2367", optarg)) != 0)
                                goto failure;
                        break;
                case 'c':
                        if ((ret = libtriplex_setparam_coerce(ctx, "mi_0145", optarg)) != 0)
                                goto failure;
                        break;
                case 'd':
                        if ((ret = libtriplex_setparam_coerce(ctx, "mi_2367", optarg)) != 0)
                                goto failure;
                        break;
                case 'J':
                        if ((ret = libtriplex_setparam_coerce(ctx, "pen-isochange", optarg)) != 0)
                                goto failure;
                        break;
                case 'K':
                        if ((ret = libtriplex_setparam_coerce(ctx, "pen-mismatch", optarg)) != 0)
                                goto failure;
                        break;
                case 'D':
                        if ((ret = libtriplex_setparam_coerce(ctx, "pen-insdel", optarg)) != 0)
                                goto failure;
                        break;
                case 'M':
                        if ((ret = libtriplex_setparam_coerce(ctx, "dtwist-limit", optarg)) != 0)
                                goto failure;
                        break;
                case 'N':
                        if ((ret = libtriplex_setparam_coerce(ctx, "score-isostay", optarg)) != 0)
                                goto failure;
                        break;
                case 'I':
                        libtriplex_setflags(ctx, LIBTRIPLEX_FLAG_IDXOFF);
                        break;
                case 'S':
                        libtriplex_setflags(ctx, LIBTRIPLEX_FLAG_CDSOFF);
                        break;
                case 'v':
                        verbose_mode = true;
                        break;
                case 'h':
                        print_usage(stdout, ctx, arg0);
                        goto cleanup_and_exit;
                case 'f': /* compatibility with older versions */
                        infile = optarg;
                        break;
                default:
                        fprintf(stderr, "[!] Unknown option `-%c'\n", opt);
                        print_usage(stderr, ctx, arg0);
                        ret = EXIT_FAILURE;
                        goto cleanup_and_exit;
                failure:
                        fprintf(stderr, "[!] An error ocured while processing option `-%c': ", opt);

                        switch(ret) {
                        case LIBTRIPLEX_ESYSTEM:
                                fprintf(stderr, "system error: %s (errno=%d)\n", strerror(errno), errno);
                                break;
                        case LIBTRIPLEX_EINVALP:
                                fprintf(stderr, "invalid parameter (probably a bug, please report this to the authors)\n");
                                print_usage(stderr, ctx, arg0);
                                break;
                        case LIBTRIPLEX_EINVALV:
                                fprintf(stderr, "invalid parameter value: \"%s\"\n", optarg);
                                print_usage(stderr, ctx, arg0);
                                break;
                        default:
                                fprintf(stderr, "unknown error (probably a bug, please report this to the authors)\n");
                                break;
                        }

                        ret = EXIT_FAILURE;
                        goto cleanup_and_exit;
                }
        }

        argc -= optind;
        argv += optind;

        /*
         * If neither `-f' or a file was specified, exit with error.
         */
        if (argc == 0) {
                if (infile == NULL) {
                        fprintf(stderr, "[!] No input file\n");
                        print_usage(stderr, ctx, arg0);
                        ret = EXIT_FAILURE;
                        goto cleanup_and_exit;
                }
        } else
                infile = argv[0];

        /*
         * Create a new GFF 3 db, with "triplex" set as the
         * source (see GFF 3 format definition).
         */
        gffres = gff3_new("triplex");

        dP("DEBUG: input=%s output=%s\n", infile, opt_outpath);

        /*
         * Execute the search
         */
        switch(ret = libtriplex_search(ctx, infile, gffres)) {
        case 0:
                break;
        case LIBTRIPLEX_ESOURCE:
                fprintf(stderr, "Can't load sequence data from \"%s\": %s\n", infile, strerror(errno));
                goto cleanup_and_exit;
        default:
                fprintf(stderr, "Search failed: ret=%d\n", ret);
                goto cleanup_and_exit;
        }

        /**
         * Process results
         */
        dP("DEBUG: writing results to %s\n", opt_outpath);
        gff3_write(gffres, opt_outpath);

cleanup_and_exit:
        /**
         * Cleanup & exit
         */
        libtriplex_fini(ctx);
        config_destroy(&cfg);

        if (gffres != NULL)
                gff3_free(gffres);
        if (opt_profile != NULL)
                free(opt_profile);
        if (opt_outpath != NULL)
                free(opt_outpath);

        return (ret);
}
