/*
 *  libtriplex.c: implementation section of the triplex search
 *             program. Searches for potential triplex sequences in DNA.
 *             can be used as a library or together with triplex.c
 *             to compile into a command-line application.
 *  Copyright (C) 2011
 *    Brno University of Technology, Faculty of Information Technology
 *    Masaryk University, Faculty of Informatics
 *
 *  Author(s): Tomas Martinek <martinto@fit.vutbr.cz>
 *             Daniel Kopecek <xkopecek@fi.muni.cz>
 *             Matej Lexa <lexa@fi.muni.cz>
 *
 *  LICENSE TERMS
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions and the following disclaimer in
 *  the documentation and/or other materials provided with the
 *  distribution.
 *  3. Neither the name of the Company nor the names of its contributors
 *  may be used to endorse or promote products derived from this
 *  software without specific prior written permission.
 *
 *  This software is provided ``as is'', and any express or implied
 *  warranties, including, but not limited to, the implied warranties of
 *  merchantability and fitness for a particular purpose are disclaimed.
 *  In no event shall the company or contributors be liable for any
 *  direct, indirect, incidental, special, exemplary, or consequential
 *  damages (including, but not limited to, procurement of substitute
 *  goods or services; loss of use, data, or profits; or business
 *  interruption) however caused and on any theory of liability, whether
 *  in contract, strict liability, or tort (including negligence or
 *  otherwise) arising in any way out of the use of this software, even
 *  if advised of the possibility of such damage.
 */

#include <stdio.h>
#include <stddef.h>
#include <string.h>
#include <stdbool.h>
#include <inttypes.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include <assert.h>

#include "libtriplex/libtriplex.h"
#include "libtriplex/gff3.h"
#include "common/helpers.h"
#include "algorithm.h"

/*
Tritype
 0: parallel triplex on sense strand
 1: parallel triplex on sense strand
 2: parallel triplex on antisense strand
 3: parallel triplex on antisense strand
 4: antiparallel triplex on sense strand
 5: antiparallel triplex on sense strand
 6: antiparallel triplex on antisense strand
 7: antiparallel triplex on antisense strand
*/

/* Lookup tables */
const int tab_score[NUM_TRI_TYPES][4][4] = {
 /*
  * Parallel
  */

 /* [W][H] */ /*  T   G   C   A */
 /*   A   */  { { 2, -9, -9, -9},
 /*   C   */    { 1, -9, -9, -9},
 /*   G   */    { 1,  1,  2, -9},
 /*   T   */    {-9,  2, -9, -9} },

 /* [H][W] */ /*  A   C   G   T */
 /*   T   */  { { 2,  1,  1, -9},
 /*   G   */    {-9, -9,  1,  2},
 /*   C   */    {-9, -9,  2, -9},
 /*   A   */    {-9, -9, -9, -9} },

 /* [W][H] */ /*  A   C   G   T */
 /*   T   */  { {-9, -9,  2, -9},
 /*   G   */    {-9,  2,  1,  1},
 /*   C   */    {-9, -9, -9,  1},
 /*   A   */    {-9, -9, -9,  2} },

 /* [H][W] */ /*  T   G   C   A */
 /*   A   */  { {-9, -9, -9, -9},
 /*   C   */    {-9,  2, -9, -9},
 /*   G   */    { 2,  1, -9, -9},
 /*   T   */    {-9,  1,  1,  2} },

 /*
  * Antiparallel
  */

 /* [W][H] */ /*  A   C   G   T */
 /*   A   */  { { 2,  1, -9,  2},
 /*   C   */    {-9, -9, -9,  1},
 /*   G   */    { 1, -9,  2, -9},
 /*   T   */    {-9, -9, -9, -9} },

 /* [H][W] */ /*  A   C   G   T */
 /*   A   */  { { 2, -9,  1, -9},
 /*   C   */    { 1, -9, -9, -9},
 /*   G   */    {-9, -9,  2, -9},
 /*   T   */    { 2,  1, -9, -9} },

 /* [W][H] */ /*  T   G   C   A */
 /*   T   */  { {-9, -9, -9, -9},
 /*   G   */    {-9,  2, -9,  1},
 /*   C   */    { 1, -9, -9, -9},
 /*   A   */    { 2, -9,  1,  2} },

 /* [H][W] */ /*  T   G   C   A */
 /*   T   */  { {-9, -9,  1,  2},
 /*   G   */    {-9,  2, -9, -9},
 /*   C   */    {-9, -9, -9,  1},
 /*   A   */    {-9,  1, -9,  2} },
};

const int tab_bound[NUM_TRI_TYPES][4][4] = {
 /*
  * Parallel
  */

 /* [W][H] */ /*  T   G   C   A */
 /*   A   */  { { 1,  0,  0,  0},
 /*   C   */    { 3,  0,  0,  0},
 /*   G   */    { 3,  3,  1,  0},
 /*   T   */    { 0,  2,  0,  0} },

 /* [H][W] */ /*  A   C   G   T */
 /*   T   */  { { 1,  3,  3,  0},
 /*   G   */    { 0,  0,  3,  2},
 /*   C   */    { 0,  0,  1,  0},
 /*   A   */    { 0,  0,  0,  0} },

 /* [W][H] */ /*  A   C   G   T */
 /*   T   */  { { 0,  0,  2,  0},
 /*   G   */    { 0,  1,  3,  3},
 /*   C   */    { 0,  0,  0,  3},
 /*   A   */    { 0,  0,  0,  1} },

 /* [H][W] */ /*  T   G   C   A */
 /* A */      { { 0,  0,  0,  0},
 /* C */        { 0,  1,  0,  0},
 /* G */        { 2,  3,  0,  0},
 /* T */        { 0,  3,  3,  1} },

 /*
  * Antiparallel
  */

 /* [W][H] */ /*  A   C   G   T */
 /*   A   */  { { 1,  1,  0,  1},
 /*   C   */    { 0,  0,  0,  2},
 /*   G   */    { 3,  0,  2,  0},
 /*   T   */    { 0,  0,  0,  0} },

 /* [H][W] */ /*  A   C   G   T */
 /*   A   */  { { 1,  0,  3,  0},
 /*   C   */    { 1,  0,  0,  0},
 /*   G   */    { 0,  0,  2,  0},
 /*   T   */    { 1,  2,  0,  0} },

 /* [W][H] */ /*  T   G   C   A */
 /*   T   */  { { 0,  0,  0,  0},
 /*   G   */    { 0,  2,  0,  3},
 /*   C   */    { 2,  0,  0,  0},
 /*   A   */    { 1,  0,  1,  1} },

 /* [H][W] */ /*  T   G   C   A */
 /*   T   */  { { 0,  0,  2,  1},
 /*   G   */    { 0,  2,  0,  0},
 /*   C   */    { 0,  0,  0,  1},
 /*   A   */    { 0,  3,  0,  1} },
};

/*
 * Triplet C1-C1-C1 angle
 */
const uint8_t tab_twist[NUM_TRI_TYPES][4][4] = {
#if !defined(OLD_MODEL)
 /*
  * Parallel
  */

 /* [W][H] */ /*    T     G    C  A */
 /*   A   */  { { 104,    0,   0, 0},
 /*   C   */    {  71,    0,   0, 0},
 /*   G   */    {  78,   75, 109, 0},
 /*   T   */    {   0,  126,   0, 0} },

 /* [H][W] */ /*    A   C    G    T */
 /*   T   */  { { 104, 71,  78,   0},
 /*   G   */    {   0,  0,  75, 126},
 /*   C   */    {   0,  0, 109,   0},
 /*   A   */    {   0,  0,   0,   0} },

 /* [W][H] */ /*   A    C    G    T */
 /*   T   */  { {  0,   0, 126,   0},
 /*   G   */    {  0, 109,  75,  78},
 /*   C   */    {  0,   0,   0,  71},
 /*   A   */    {  0,   0,   0, 104} },

 /* [H][W] */ /*    T    G   C    A */
 /* A */      { {   0,   0,  0,   0},
 /* C */        {   0, 109,  0,   0},
 /* G */        { 126,  75,  0,   0},
 /* T */        {   0,  78, 71, 104} },

 /*
  * Antiparallel
  */

 /* [W][H] */ /*    A    C   G    T */
 /*   A   */  { {  77,  72,  0,  72},
 /*   C   */    {   0,   0,  0,  94},
 /*   G   */    { 126,   0, 94,   0},
 /*   T   */    {   0,   0,  0,   0} },

 /* [H][W] */ /*  A    C     G   T */
 /*   A   */  { { 77,  0,  126,  0},
 /*   C   */    { 72,  0,    0,  0},
 /*   G   */    {  0,  0,   94,  0},
 /*   T   */    { 72, 94,    0,  0} },

 /* [W][H] */ /*   T   G   C     A */
 /*   T   */  { {  0,  0,  0,    0},
 /*   G   */    {  0, 94,  0,  126},
 /*   C   */    { 94,  0,  0,    0},
 /*   A   */    { 72,  0, 72,   77} },

 /* [H][W] */ /*  T     G   C    A */
 /*   T   */  { { 0,    0, 94,  72},
 /*   G   */    { 0,   94,  0,   0},
 /*   C   */    { 0,    0,  0,  72},
 /*   A   */    { 0,  126,  0,  77} },

#else
           /* A   C   G   T  */
 /* T */   { {71, 99, 34, 71},
 /* G */     {26, 56, 24, 31},
 /* C */     {83, 93, 49, 59},
 /* A */     {33, 45, 23, 53} },

           /* A   C   G   T  */
 /* T */   { {71, 26, 83, 33},
 /* G */     {99, 56, 93, 45},
 /* C */     {34, 24, 49, 23},
 /* A */     {71, 31, 59, 53} },

           /* T   G   C   A  */
 /* A */   { {53, 23, 45, 33},
 /* C */     {59, 49, 93, 83},
 /* G */     {31, 24, 56, 26},
 /* T */     {71, 34, 99, 71} },

           /* T   G   C   A  */
 /* A */   { {53, 59, 31, 71},
 /* C */     {23, 49, 24, 34},
 /* G */     {45, 93, 56, 99},
 /* T */     {33, 83, 26, 71} },

           /*  A   C    G   T  */
 /* A */   { { 81, 126,  4, 103},
 /* C */     { 64,  80, 43,  45},
 /* G */     {118, 126, 75, 116},
 /* T */     { 59,  70, 40,  51} },

           /*  A   C   G    T  */
 /* A */   { { 81, 64, 118, 59},
 /* C */     {126, 80, 126, 70},
 /* G */     {  4, 43,  75, 40},
 /* T */     {103, 45, 116, 51} },

           /*  T   G    C    A  */
 /* T */   { { 51, 40,  70,  59},
 /* G */     {116, 75, 126, 118},
 /* C */     { 45, 43,  80,  64},
 /* A */     {103,  4, 126,  81} },

           /* T   G    C   A  */
 /* T */   { {51, 116, 45, 103},
 /* G */     {40,  75, 43,   4},
 /* C */     {70, 126, 80, 126},
 /* A */     {59, 118, 64,  81} }
#endif
};


/* Default values */
// double LAMBDA[NUM_TRI_TYPES] = {0.67, 0.67, 0.71, 0.71, 0.67, 0.67, 0.71, 0.71};
//double MI[NUM_TRI_TYPES]     = {6.05, 6.05, 5.88, 5.88, 6.05, 6.05, 5.88, 5.88};

/* E. coli values */
/* double LAMBDA[NUM_TRI_TYPES] = {0.65, 0.65, 0.74, 0.74, 0.65, 0.65, 0.74, 0.74}; */
/* double MI[NUM_TRI_TYPES]     = {6.09, 6.09, 5.78, 5.78, 6.09, 6.09, 5.78, 5.78}; */

/* Human */
//double LAMBDA[NUM_TRI_TYPES] = {0.68, 0.68, 0.68, 0.68, 0.68, 0.68, 0.68, 0.68};
//double MI[NUM_TRI_TYPES]     = {5.98, 5.98, 5.97, 5.97, 5.98, 5.98, 5.97, 5.97};

/*
 * --------------------------------------------------------------------------
 *                          Procedures and function
 * --------------------------------------------------------------------------
 */
static double p_function(int score, int tri_type, double lambda[NUM_TRI_TYPES], double mi[NUM_TRI_TYPES]) {
        return 1 - exp(-exp(-lambda[tri_type] * (score - mi[tri_type])));
}

static double e_value(int score, int tri_type, double lambda[NUM_TRI_TYPES], double mi[NUM_TRI_TYPES]) {
        return p_function(score, tri_type, lambda, mi);
}

static double p_value(int score, int tri_type, double lambda[NUM_TRI_TYPES], double mi[NUM_TRI_TYPES]) {
        return 1 - exp(-p_function(score, tri_type, lambda, mi));
}

/**
 * Calcutates the length of a path in DP matrix
 */
static int get_length(int start_antidiag, int end_antidiag, int start_diag, int end_diag)
{
        register int length;

        length = (end_antidiag - start_antidiag + 2)/2;
        /* diagonal odd to even - increment length */
        if (((start_diag % 2) == 1) && ((end_diag % 2) == 0))
                length++;
        if (length == 0)
                length = 1;

        return length;
}

/**
 * Application of DP rule
 */
static t_diag get_max_score(char a, char b, t_diag *dl, t_diag *d, t_diag *dr, int diag, int antidiag, int tri_type, int max_loop, sc_scheme_t *sc)
{
        t_diag res;
        register int16_t incscore, mm_score;

        /* --------------------------------------------------------------- */
        /* match or mismatch score computation  */
        incscore = tab_score[tri_type][a][b];

        /*
         * a trick to have zero or negative scores
         * for low quality triplets, changed all
         * incscore comparisons to -9
         */
        if (incscore > -9) {
                /*
                 * Match
                 */
                mm_score = d->score + incscore;

                /*
                 * Penalize isogroup change and large angle differences
                 */
                if ((d->dpst & DPST_DPMASK) == DP_MATCH) {
                        if ((tab_bound[tri_type][a][b] != d->bound) && (abs(tab_twist[tri_type][a][b] - d->twist) > sc->dtwist_limit ) && (abs(tab_twist[tri_type][a][b] - d->twist + d->dtwist) > sc->dtwist_limit))
                                mm_score -= sc->pen_isochange;
                        else
                                mm_score += sc->score_isostay;
                }
        } else {
                /*
                 * Mismatch
                 */
                mm_score = d->score - sc->pen_mismatch;
        }

        /*
         * Match/Mismatch is better
         */
        if ((mm_score >= dl->score - sc->pen_insdel) &&
            (mm_score >= dr->score - sc->pen_insdel))
        {
                res = *d;
                res.score = mm_score;
                res.dpst  = (res.dpst & DPST_STMASK) | DP_MISMATCH;

                /* match */
                if (incscore > -9) {
                        res.dpst    = (res.dpst & DPST_STMASK) | DP_MATCH;
                        res.bound   = tab_bound[tri_type][a][b];
                        res.twist   = tab_twist[tri_type][a][b];
                        res.dtwist  = res.twist - d->twist;

                        if (mm_score >= res.max_score) {
                                res.max_score = mm_score;
                                res.max_score_pos.diag = diag;
                                res.max_score_pos.antidiag = antidiag;
                                res.max_indels = res.indels;
                        }
                }
        }
        /* --------------------------------------------------------------- */
        /* insetion or deletion */
        else {
                /* get from left diagonal */
                if (dl->score > dr->score) {
                        res = *dl;
                        res.score = dl->score - sc->pen_insdel;
                        res.dpst = (res.dpst & DPST_STMASK) | DP_LEFT;
                }
                /* get from right diagonal */
                else {
                        res = *dr;
                        res.score = dr->score - sc->pen_insdel;
                        res.dpst = (res.dpst & DPST_STMASK) | DP_RIGHT;
                }

                res.indels++;
        }
        /* --------------------------------------------------------------- */
        /* local alignment - only for loop */
        if ((res.score < 0) && (antidiag <= max_loop)) {
                res.score = 0;
                res.max_score = 0;
                res.start.diag = diag;
                res.start.antidiag = antidiag;
                res.max_score_pos.diag = diag;
                res.max_score_pos.antidiag = antidiag;
                res.indels = 0;
                res.max_indels = 0;
        }

        /* --------------------------------------------------------------- */
        return res;
}

static void export_data_gff3(gff3_t *gff, const char *seq_id, t_diag *diag, bool exported, const uint8_t *data, int tri_type, double p_val, uint64_t offset)
{
        gff3_rec_t gffrec;
        uint64_t gap_start, gap_end;
        int length;
        char numbuf[32+1];

        (void)data;

        gff3_rec_prepare(&gffrec, 6);

        gffrec.end   = offset + (uint64_t)((diag->max_score_pos.diag + diag->max_score_pos.antidiag - 1)/2);
        gffrec.start = offset + (uint64_t)(gffrec.end - diag->max_score_pos.antidiag);

        ++gffrec.end;
        ++gffrec.start;

        gffrec.seq_id = (char *)seq_id;
        gffrec.type   = "match";

        gap_end   = offset + (diag->start.diag + diag->start.antidiag - 1)/2;
        gap_start = offset + (gap_end - diag->start.antidiag);

        ++gap_end;
        ++gap_start;

        length = get_length(diag->start.antidiag, diag->max_score_pos.antidiag,
                            diag->start.diag,     diag->max_score_pos.diag);

        gffrec.score = (double)((int)diag->max_score);

        switch (tri_type) {
        case 0: case 1:
        case 4: case 5:
                gffrec.strand = '+';
                break;
        case 2: case 3:
        case 6: case 7:
                gffrec.strand = '-';
                break;
        }

        gffrec.phase = -1;

        gff3_rec_addtag(&gffrec, "exported", strdup(exported ? "1" : "0"));

        snprintf(numbuf, sizeof numbuf - 1, "%"PRIu64, gap_start);
        gff3_rec_addtag(&gffrec, "gap_start", strdup(numbuf));

        snprintf(numbuf, sizeof numbuf - 1, "%"PRIu64, gap_end);
        gff3_rec_addtag(&gffrec, "gap_end",   strdup(numbuf));

        snprintf(numbuf, sizeof numbuf - 1, "%d", length);
        gff3_rec_addtag(&gffrec, "length",    strdup(numbuf));

        snprintf(numbuf, sizeof numbuf - 1, "%e", p_val);
        gff3_rec_addtag(&gffrec, "p_val",     strdup(numbuf));

        snprintf(numbuf, sizeof numbuf - 1, "%"PRIu16, diag->max_indels);
        gff3_rec_addtag(&gffrec, "indels",    strdup(numbuf));

        snprintf(numbuf, sizeof numbuf - 1, "%d", tri_type);
        gff3_rec_addtag(&gffrec, "type", strdup(numbuf));

        gff3_rec_add(gff, &gffrec);
        return;
}

int triplex_search(triplexCTX *ctx, const char *buffer_id, const uint8_t *buffer, size_t buflen, uint64_t offset, int tri_type, gff3_t *result)
{
        register size_t i, d, x;
        register int c, length, num;
        int exp_num, nexp_num;

        /*
         * Variables for storing the algorithm parameters
         */
        int max_steps;
        int min_length;
        int min_score;
        int min_loop;
        int max_loop;

        double p_val, pv;

        double lm_0145, lm_2367;
        double lambda[NUM_TRI_TYPES];

        double mi_0145, mi_2367;
        double mi[NUM_TRI_TYPES];

        t_diag *diag;
        sc_scheme_t sc;

        assert(ctx != NULL);
        assert(result != NULL);
        assert(buffer != NULL);
        assert(buffer_id != NULL);

        /*
         * Get all needed parameter values
         */
        libtriplex_getparam(ctx, "max-steps",       TRIPLEX_PARAM_INT, &max_steps);
        libtriplex_getparam(ctx, "min-length",      TRIPLEX_PARAM_INT, &min_length);
        libtriplex_getparam(ctx, "min-score",       TRIPLEX_PARAM_INT, &min_score);
        libtriplex_getparam(ctx, "min-loop-length", TRIPLEX_PARAM_INT, &min_loop);
        libtriplex_getparam(ctx, "max-loop-length", TRIPLEX_PARAM_INT, &max_loop);

        libtriplex_getparam(ctx, "p-value",     TRIPLEX_PARAM_FLOAT, &p_val);

        libtriplex_getparam(ctx, "lambda_0145", TRIPLEX_PARAM_FLOAT, &lm_0145);
        libtriplex_getparam(ctx, "lambda_2367", TRIPLEX_PARAM_FLOAT, &lm_2367);

        lambda[0] = lambda[1] = lambda[4] = lambda[5] = lm_0145;
        lambda[2] = lambda[3] = lambda[6] = lambda[7] = lm_2367;

        libtriplex_getparam(ctx, "mi_0145", TRIPLEX_PARAM_FLOAT, &mi_0145);
        libtriplex_getparam(ctx, "mi_2367", TRIPLEX_PARAM_FLOAT, &mi_2367);

        mi[0] = mi[1] = mi[4] = mi[5] = mi_0145;
        mi[2] = mi[3] = mi[6] = mi[7] = mi_2367;

        libtriplex_getparam(ctx, "pen-isochange", TRIPLEX_PARAM_INT, &sc.pen_isochange);
        libtriplex_getparam(ctx, "pen-mismatch",  TRIPLEX_PARAM_INT, &sc.pen_mismatch);
        libtriplex_getparam(ctx, "pen-insdel",    TRIPLEX_PARAM_INT, &sc.pen_insdel);
        libtriplex_getparam(ctx, "dtwist-limit",  TRIPLEX_PARAM_INT, &sc.dtwist_limit);
        libtriplex_getparam(ctx, "score-isostay", TRIPLEX_PARAM_INT, &sc.score_isostay);

        dP("   triplex type: %d\n", tri_type);
        dP("      max steps: %d\n", max_steps);
        dP("     min length: %d\n", min_length);
        dP("      min score: %d\n", min_score);
        dP("min loop length: %d\n", min_loop);
        dP("max loop length: %d\n", max_loop);
        dP("        p value: %f\n", p_val);
        dP("            l 1: %f\n", lm_0145);
        dP("            l 2: %f\n", lm_2367);
        dP("            m 1: %f\n", mi_0145);
        dP("            m 2: %f\n", mi_2367);

        /*
         * Allocate and initialize the diag array
         */
        dP("DEBUG: Allocating %zu bytes for diagonals (%zu sz/item)\n",
               sizeof(t_diag) * 2 * buflen, sizeof(t_diag));

        diag = alloc_array(t_diag, 2 * buflen);

        if (diag == NULL)
                return (LIBTRIPLEX_ENOMEM);

        /*
         * Initialize the diagonals
         */
        for (i = 0;  i < 2 * buflen; ++i) {
                diag[i].score     = 0;
                diag[i].max_score = 0;
                diag[i].bound     = 0;
                diag[i].twist     = 90;
                diag[i].dtwist    = 0;

                diag[i].start.diag     = i;
                diag[i].start.antidiag = (((min_loop + i) % 2) == 0) ? min_loop + 1 : min_loop + 2;

                diag[i].max_score_pos.diag     = diag[i].start.diag;
                diag[i].max_score_pos.antidiag = diag[i].start.antidiag;

                diag[i].indels     = 0;
                diag[i].max_indels = 0;
                diag[i].dpst       = DP_MISMATCH|STAT_NONE;

                /* minloop    diag       antidiag */
                /* sude       sude         +0     */
                /* sude       liche        +1 */
                /* liche      sude         +1 */
                /* liche      liche        +0 */
        }

        /*
         * Main computation loop
         */
        exp_num = 0;

        for (x = min_loop + 1; x <= (size_t)(2 * max_steps) && x < buflen; ++x) {
                d = x + 1;

                for (i = x; i < buflen; i++) {
                        /*
                         * max score and length calcualtion
                         */
                        diag[d] = get_max_score(buffer[i], buffer[i - x], diag + d - 1, diag + d, diag + d + 1,
                                                d, x, tri_type, max_loop, &sc);

                        length  = get_length(diag[d].start.antidiag, diag[d].max_score_pos.antidiag,
                                             diag[d].start.diag,     diag[d].max_score_pos.diag);

                        diag[d].dpst = (length >= min_length) ?
                                diag[d].dpst | STAT_MINLEN : diag[d].dpst & (~STAT_MINLEN);

                        /*
                         * actual score satisfies the required quality
                         */
                        if (diag[d].score >= min_score) {
                                diag[d].dpst |= STAT_QUALITY;

                                /*
                                 * if triplex can not continue, then export
                                 */
                                if ((diag[d].dpst & STAT_MINLEN) &&
                                    (d == (x+1) || d == (2 * buflen - x - 1)))
                                {
                                        pv = p_value(diag[d].max_score, tri_type, lambda, mi);

                                        exp_num++;
                                        diag[d].dpst = (diag[d].dpst & DPST_DPMASK) | STAT_EXPORT;

                                        if (pv < p_val)
                                                export_data_gff3(result, buffer_id, diag + d, true, buffer, tri_type, pv, offset);
                                }
                        } else {/* actual score does not satisfy the required quality */

                                /*
                                 * if quality requirement was satisfied
                                 * in previous step, then export
                                 */
                                if((!(diag[d-1].dpst & STAT_QUALITY)) &&
                                   (!(diag[d+1].dpst & STAT_QUALITY)) &&
                                   ( (diag[ d ].dpst & STAT_QUALITY)) &&
                                   ( (diag[ d ].dpst & STAT_MINLEN)))
                                {
                                        pv = p_value(diag[d].max_score, tri_type, lambda, mi);
                                        exp_num++;
                                        diag[d].dpst = (diag[d].dpst & DPST_DPMASK) | STAT_EXPORT;

                                        if (pv < p_val)
                                                export_data_gff3(result, buffer_id, diag + d, true, buffer, tri_type, pv, offset);

                                        diag[d].max_score = 0;
                                } else {
                                        diag[d].dpst = (diag[d].dpst & DPST_DPMASK) | STAT_NONE;
                                }
                        }

                        d += 2;
                }
        }

        nexp_num = 0;

        for (i = 1; (size_t)i < (2 * buflen); i++) {
                if ((diag[i].dpst & STAT_QUALITY) &&
                    (diag[i].dpst & STAT_MINLEN))
                {
                        pv = p_value(diag[i].max_score, tri_type, lambda, mi);

                        if(pv < p_val)
                                export_data_gff3(result, buffer_id, diag + i, false, buffer, tri_type, pv, offset);

                        nexp_num++;
                }
        }

        free(diag);

        return (0);
}
